package pc;

import java.util.ArrayList;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingDeque;

// needed parts of java.util.concurrent.Queue
interface ProducerConsumerQueue<T> {
    void add(T e);
    T take() throws InterruptedException;
}

class JavaUtilConcurrentProducerConsumerQueue<T> implements ProducerConsumerQueue<T> {
    private final BlockingQueue<T> data = new LinkedBlockingDeque<>();

    public void add(T element) {
        data.add(element);
    }

    @Override
    public T take() throws InterruptedException {
        return data.take();
    }
}

class MyProducerConsumerQueue<T> implements ProducerConsumerQueue<T> {
    private final ArrayList<T> data = new ArrayList<>();

    public void add(T element) {
        synchronized (data) {
            data.add(element);
            data.notifyAll();
        }
    }

    public T take() {
        synchronized (data) {
            // while needed for spourious wakeups https://en.wikipedia.org/wiki/Spurious_wakeup
            while (data.isEmpty()) {
                try {
                    data.wait();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
            return data.remove(0);
        }
    }
}

class Consumer<U> extends Thread implements java.util.function.Consumer<U> {
    private final ProducerConsumerQueue<U> queue;

    public Consumer(ProducerConsumerQueue<U> queue) {
        this.queue = queue;
    }

    @Override
    public void accept(U integer) {
        queue.add(integer);
    }

    @Override
    public java.util.function.Consumer<U> andThen(java.util.function.Consumer<? super U> after) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void run() {
        for (int i = 0; i < 100; ++i) {
            try {
                queue.take();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }
}
